module overlap_m

  implicit none

contains

  subroutine overlap(flow, e_overestim, k, delta, j)

    ! This procedure finds edges between flow(j - delta) and flow(j),
    ! corresponding to date indices k - delta and k. It writes these
    ! edges. It updates flow(j - delta)%list%delta_out and
    ! flow(j)%list%delta_in.

    ! Libraries:
    use contour_531, only: polyline
    use gpc_f, only: gpc_polygon_clip_f, GPC_INT, polygon
    use jumble, only: twopi

    use candidate_overlap_m, only: candidate_overlap
    use config_graph_m, only: dist_lim, min_inters
    use derived_types, only: snapshot
    use read_grid_m, only: nlon, nlat, periodic
    use spher_polygon_area_m, only: spher_polygon_area
    use spher_polyline_area_m, only: spher_polyline_area
    use unit_edge_m, only: unit_edge

    type(snapshot), intent(inout):: flow(:) ! (max_delta + 1)
    integer, intent(in):: e_overestim
    ! over-estimation of the number of eddies at each date

    integer, intent(in):: k ! date index
    integer, intent(in):: delta ! between 1 and max_delta

    integer, intent(in):: j
    ! position in time window, between 2 and max_delta + 1

    ! Local:

    integer i1 ! eddy index of predecessor
    integer i2 ! eddy index of successor
    integer l, n_select
    type(polyline) polyline_1, polyline_2
    type(polygon) res_pol

    integer, allocatable:: selection(:)
    ! identifying numbers of a selection of extrema

    ! Window around each extremum:
    integer llc(2) ! indices in global grid of lower left corner
    integer urc(2) ! indices in global grid of upper right corner

    !-----------------------------------------------------------------------

    loop_i1: do i1 = 1, flow(j - delta)%number_extr
       associate (e1 => flow(j - delta)%list(i1))
         test_valid: if (e1%valid) then
            ! Define the geographical window around each eddy extremum:

            llc = e1%extr%coord_proj - dist_lim
            urc = e1%extr%coord_proj + dist_lim

            llc(2) = max(llc(2), 1)
            urc(2) = min(urc(2), nlat)

            if (.not. periodic) then
               llc(1) = max(llc(1), 1)
               urc(1) = min(urc(1), nlon)
            end if

            ! Pre-select potential successors:
            selection = candidate_overlap(flow(j)%extr_map(llc(1):urc(1), &
                 llc(2):urc(2)), flow(j)%list, e1%delta_out, delta)

            n_select = size(selection)

            if (n_select /= 0) then
               if (e1%speed_cont%n_points /= 0) then
                  polyline_1 = e1%speed_cont%polyline
               else
                  polyline_1 = e1%out_cont%polyline
               end if
            end if

            DO l = 1, n_select
               i2 = selection(l)
               associate (e2 => flow(j)%list(i2))
                 ! Assertion: {e1%delta_out >= delta .or. e2%delta_in
                 ! >= delta}

                 if (e2%speed_cont%n_points /= 0) then
                    polyline_2 = e2%speed_cont%polyline
                 else
                    polyline_2 = e2%out_cont%polyline
                 end if

                 ! Shift the longitudes of polyline_2 to values close to the
                 ! longitude of extremum i1:
                 polyline_2%points(1, :) = polyline_2%points(1, :) &
                      + floor((e1%extr%coord(1) - e2%extr%coord(1)) / twopi &
                      + 0.5) * twopi

                 call gpc_polygon_clip_f(GPC_INT, polygon(nparts = 1, &
                      part = [polyline_1], hole = [.false.]), &
                      polygon(nparts = 1, part = [polyline_2], &
                      hole = [.false.]), res_pol)

                 if (res_pol%nparts /= 0) then
                    ! polyline_1 and polyline_2 overlap
                    if (spher_polygon_area(res_pol) >= min_inters &
                         * min(abs(spher_polyline_area(polyline_1)), &
                         abs(spher_polyline_area(polyline_2)))) then
                       write(unit_edge, fmt = *) &
                            (k - delta) * e_overestim + i1, k * e_overestim + i2
                       e1%delta_out = min(e1%delta_out, delta)
                       e2%delta_in = min(e2%delta_in, delta)
                    end if
                 end if
               end associate
            end DO
         end if test_valid
       end associate
    end do loop_i1

  end subroutine overlap

end module overlap_m
